#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# @Author : 陈坤泽
# @Email  : 877362867@qq.com
# @Date   : 2020/03/16 09:19


"""一些python通用功能的性能测试
    虽然其实大部分场合其实都是相通的
    有时候test测试代码，其实也是演示如何使用的demo

demo：示例代码，注重演示
debug：调试代码，注重分析自己代码功能是否有bug
test：测试代码，注重分析功能稳定性
perf：性能测试，注重分析代码的运行效率
"""

from pyxllib.xl import *

____stdlib = """
标准库相关
"""


def demo_system():
    """主要是测试一些系统变量值，顺便再演示一次Timer用法"""

    def demo_pc_messages():
        """演示如何获取当前操作系统的PC环境数据"""
        # fqdn：fully qualified domain name
        print('1、socket.getfqdn() :', socket.getfqdn())  # 完全限定域名，可以理解成pcname，计算机名
        # 注意py的很多标准库功能本来就已经处理了不同平台的问题，尽量用标准库而不是自己用sys.platform作分支处理
        print('2、sys.platform     :', sys.platform)  # 运行平台，一般是win32和linux
        li = os.getenv('PATH').split(os.path.pathsep)  # 环境变量名PATH，win中不区分大小写，linux中区分大小写必须写成PATH
        print("3、os.getenv('PATH'):", f'数量={len(li)},', pprint.pformat(li, 4))

    def demo_executable_messages():
        """演示如何获取被执行程序相关的数据"""
        print('1、sys.path      :', f'数量={len(sys.path)},', pprint.pformat(sys.path, 4))  # import绝对位置包的搜索路径
        print('2、sys.executable:', sys.executable)  # 当前被执行脚本位置
        print('3、sys.version   :', sys.version)  # python的版本
        print('4、os.getcwd()   :', os.getcwd())  # 获得当前工作目录
        print('5、gettempdir()  :', tempfile.gettempdir())  # 临时文件夹位置

    timer = Timer('demo_system')
    print('>>> demo_pc_messages()')
    demo_pc_messages()
    print('>>> demo_executable_messages()')
    demo_executable_messages()
    timer.stop_and_report()


def test_re():
    """ 正则re模块相关功能测试
    """
    # 190103周四
    # py的正则[ ]语法，可以用连字符-匹配一个区间内的字符，
    # 例如数字0-9（你是不是蠢，不会用\d么），还有a-z、A-Z（\w），甚至①-⑩，但是一-十，注意'四'匹配不到
    dprint(re.sub(r'[一-十]', '', '一二三四五六七八九十'))
    # [05]demolib.py/98: re.sub(r'[一-十]', '', '一二三四五六七八九十')<str>='四'

    # 200319周四14:11，匹配顺序与内容有关，先出现的先匹配，而与正则里or语法参数顺序无关
    print(re.findall(r'(<(a|b)>.*?</\2>)', '<a><b></b></a>'))
    print(re.findall(r'(<(b|a)>.*?</\2>)', '<a><b></b></a>'))
    # 结果都是： [('<a><b></b></a>', 'a')]
    # TODO 200323周一17:22，其实是模式不够复杂，在特殊场景下，可选条件的前后顺序是有影响的


def perf_concurrent():
    import time
    import concurrent.futures

    def func():
        s = 0
        for i in range(1000):
            for j in range(1000):
                s += j ** 5
        return s

    start = time.time()
    for i in range(5):
        func()
    print(f'单线程 During Time: {time.time() - start:.3f} s')

    start = time.time()
    executor = concurrent.futures.ThreadPoolExecutor(4)
    for i in range(5):
        executor.submit(func)
    executor.shutdown()
    print(f'多线程 During Time: {time.time() - start:.3f} s')


____pyxllib = """
pyxllib库相关
"""


def demo_timer():
    """该函数也可以用来测电脑性能
    代码中附带的示例结果是我在自己小米笔记本上的测试结果
    Intel（R） Core（TM） i7-10510U CPU@ 1.80GHz 2.30 GHz，15G 64位
    """
    import math
    import numpy

    print('1、普通用法（循环5*1000万次用时）')
    timer = Timer('循环')
    timer.start()
    for _ in range(5):
        for _ in range(10 ** 7):
            pass
    timer.stop()
    timer.report()
    # 循环 用时: 0.727s

    print('2、循环多轮计时')
    timer = Timer('自己算均值标准差耗时')

    # 数据量=200是大概的临界值，往下自己算快，往上用numpy算快
    # 临界量时，每万次计时需要0.45秒。其实整体都很快影响不大，所以Timer最终统一采用numpy来运算。
    data = list(range(10)) * 20

    for _ in range(5):
        timer.start()  # 必须明确指定每次的 开始、结束 时间
        for _ in range(10 ** 4):
            n, sum_ = len(data), sum(data)
            mean1 = sum_ / n
            std1 = math.sqrt((sum([(x - mean1) ** 2 for x in data]) / n))
        timer.stop()  # 每轮结束时标记
    timer.report()
    # 自己算均值标准差耗时 总耗时: 2.214s	均值标准差: 0.443±0.008s	总数: 5	最小值: 0.435s	最大值: 0.459s
    dprint(mean1, std1)
    # [05]timer.py/97: mean1<float>=4.5    std1<float>=2.8722813232690143

    print('3、with上下文用法')
    with Timer('使用numpy算均值标准差耗时') as t:
        for _ in range(5):
            t.start()
            for _ in range(10 ** 4):
                mean2, std2 = numpy.mean(data), numpy.std(data)
            t.stop()
    # 主要就是结束会自动report，其他没什么太大差别
    # 使用numpy算均值标准差耗时 总耗时: 2.282s	均值标准差: 0.456±0.015s	总数: 5	最小值: 0.442s	最大值: 0.483s
    dprint(mean2, std2)
    # [05]timer.py/109: mean2<numpy.float64>=4.5    std2<numpy.float64>=2.8722813232690143

    print('4、可以配合dformat输出定位信息')
    with Timer(dformat()) as t:
        for _ in range(5):
            t.start()
            for _ in range(10 ** 6):
                pass
            t.stop()
    # [04]timer.py/113:      总耗时: 0.096s	均值标准差: 0.019±0.002s	总数: 5	最小值: 0.018s	最大值: 0.023s


def demo_dprint():
    """这里演示dprint常用功能
    """
    # 1 查看程序是否运行到某个位置
    dprint()
    # [05]dprint.py/169:      意思：这是堆栈的第5层，所运行的位置是 dprint.py文件的第169行

    # 2 查看变量、表达式的 '<类型>' 和 ':值'
    a, b, s = 1, 2, 'ab'
    dprint(a, b, a ^ b, s * 2)
    # [05]dprint.py/174: a<int>=1    b<int>=2    a ^ b<int>=3    s*2<str>='abab'

    # 3 异常警告
    b = 0
    if b:
        c = a / b
    else:
        c = 0
        dprint(a, b, c)  # b=0不能作为除数，c默认值暂按0处理
    # [05]dprint.py/183: a<int>=1    b<int>=0    c<int>=0    # b=0不能作为除数，c默认值暂按0处理

    # 4 如果想在其他地方使用dprint的格式内容，可以调底层dformat函数实现
    with TicToc(dformat(fmt='[{depth:02}]{fullfilename}/{lineno}: {argmsg}')):
        for _ in range(10 ** 7):
            pass
    # [04]D:\slns\pyxllib\pyxllib\debug\pupil.py/187:  0.173 秒.


def _test_getfile_speed():
    """
    遍历D盘所有文件(205066个) 用时0.65秒
    遍历D盘所有tex文件(7796个) 用时0.95秒
    有筛选遍历D盘所有文件(193161个) 用时1.19秒
    有筛选遍历D盘所有tex文件(4464个) 用时1.22秒
        + EnsureContent： 3.18秒，用list存储所有文本要 310 MB 开销，转str拼接差不多也是这个值
        + re.sub(r'\$.*?\$', r'', s)： 4.48秒
    """
    timer = Timer(start_now=True)
    ls = list(getfiles(r'D:\\'))
    timer.stop_and_report(f'遍历D盘所有文件({len(ls)}个)')

    timer = Timer(start_now=True)
    ls = list(getfiles(r'D:\\', '.tex'))
    timer.stop_and_report(f'遍历D盘所有tex文件({len(ls)}个)')

    timer = Timer(start_now=True)
    ls = list(mygetfiles(r'D:\\'))
    timer.stop_and_report(f'有筛选遍历D盘所有文件({len(ls)}个)')

    timer = Timer(start_now=True)
    ls = list(mygetfiles(r'D:\\', '.tex'))
    timer.stop_and_report(f'有筛选遍历D盘所有tex文件({len(ls)}个)')


____perf = """
"""



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# @Author : 陈坤泽
# @Email  : 877362867@qq.com
# @Date   : 2021/06/03 23:28

import datetime
import re

import arrow

from humanfriendly import parse_date


class Datetime(arrow.Arrow):
    r"""时间戳类

    包含功能：
        各种日期格式 -> 标准日期格式 -> 各种类型的输出格式
        以及日期间的运算

    TODO 这个库的初始化接口写的太灵活了，需要一定的限制，否则会造成使用者的模糊
    """

    def __init__(self, *argv, fold=0):
        r"""超高智能日期识别，能处理常见的各种输入类型来初始化一个标准时间戳值

        :param fold: 暂时也没研究这个参数有什么用，就是为了兼容性先挂着

        # 建议使用 Datetime.now()
        >> Datetime()  # 没有参数的时候，初始化为当前时间
        <Datetime [2020-03-22T01:00:07.035481+00:00]>

        # 正常初始化方法
        >>> Datetime(2017, 2, 5, 8, 26, 58, 861782)  # 标准的时间初始化方法
        <Datetime [2017-02-05T08:26:58.861782+08:00]>
        >>> Datetime(2017)  # 可以省略月、日等值
        <Datetime [2017-01-01T00:00:00+08:00]>

        # 建议使用 Datetime.strptime
        >>> Datetime('2019年3月6日', '%Y年%m月%d日')  # 指定格式
        <Datetime [2019-03-06T00:00:00+08:00]>
        >>> Datetime('w200301周日', 'w%y%m%d周日')  # 周日必须写全，有缺失会报ValueError
        <Datetime [2020-03-01T00:00:00+08:00]>

        >>> Datetime(180213)
        <Datetime [2018-02-13T00:00:00+08:00]>
        >>> Datetime('180213')
        <Datetime [2018-02-13T00:00:00+08:00]>

        >>> Datetime('2015-06-15_22-19-01_HDR.jpg')
        <Datetime [2015-06-15T22:19:01+08:00]>
        >>> Datetime('IMG_20150615_2219011234_HDR.jpg')
        <Datetime [2015-06-15T22:19:01.001234+08:00]>
        >>> Datetime('_2015.6.15_22:19:02')
        <Datetime [2015-06-15T22:19:02+08:00]>
        """
        dt = None
        # 1 没有参数则默认当前运行时间
        if not argv:
            dt = datetime.datetime.now()
        if not dt and isinstance(argv[0], datetime.date):
            dt = datetime.datetime(argv[0].year, argv[0].month, argv[0].day)
        if not dt and isinstance(argv[0], (datetime.datetime, arrow.Arrow, Datetime)):
            dt = argv[0]
        if not dt and isinstance(argv[0], float):
            dt = datetime.datetime.fromtimestamp(argv[0])
        # 2 优先尝试用标准的datetime初始化方法
        if not dt:
            dt = Datetime._datetime(argv)
        # 3 如果上述解析不了，且argv恰好为两个参数，则判断为使用strptime初始化
        if not dt and len(argv) == 2:
            dt = datetime.datetime.strptime(str(argv[0]), argv[1])
        # 4 判断是否我个人特用的六位日期标记
        if not dt:
            dt = Datetime._six_digits_date(argv[0])
        # 5 如果仍然解析不了，开始使用一个智能推导算法
        if not dt:
            dt = Datetime._parse_time_string(argv[0])
        # 6 最后任何解析方案都失败，则报错
        if not dt:
            print(f'无法解析输入的时间标记 argv: {argv}，现重置为当前时间点')
            dt = datetime.datetime.now()

        super().__init__(dt.year, dt.month, dt.day,
                         dt.hour, dt.minute, dt.second, dt.microsecond,
                         'local', fold=getattr(dt, 'fold', 0))

    @classmethod
    def strptime(cls, data_strnig, format=None):
        return cls(*parse_date(data_strnig))

    @staticmethod
    def _datetime(argv):
        args, n = list(argv), len(argv)
        if n < 3:  # 若没填写月、日，默认1月、1日
            args = args + [1] * (3 - n)
        try:
            return datetime.datetime(*args)
        except (ValueError, TypeError) as e:
            return None

    @staticmethod
    def _six_digits_date(s):
        """主要是我个人常用的日期标注格式
        """
        s, dt = str(s), None
        if re.match(r'\d{6}$', s):
            year = int(s[:2])
            year = 2000 + year if year < 50 else 1900 + year  # 小于50默认是20xx年，否则认为是19xx年
            dt = Datetime._datetime([year, int(s[2:4]), int(s[4:])])
        return dt

    @staticmethod
    def _parse_time_string(s):
        r"""只对2000~2099年的时间点有效
        """
        data, break_flag = [], False

        def parse(pattern, left=None, right=None):
            """通用底层解析器"""
            nonlocal break_flag, s
            if break_flag: return

            m = re.search(pattern, s)
            if m:
                d = int(m.group())
                if left and d < left:
                    break_flag = True
                elif right and d > right:
                    break_flag = True
                else:
                    data.append(d)
                    s = s[m.end():]
            else:
                break_flag = True

        parse(r'20\d{2}')
        parse(r'\d\d?', 1, 12)  # 有连续两个数组就获得两个，否则获得一个也行
        parse(r'\d\d?', 1, 31)  # 这样其实不严谨，有的月份不到31天，不过反正_datetime会返回None代表处理失败的
        parse(r'\d\d?', 0, 23)
        parse(r'\d\d?', 0, 59)
        parse(r'\d\d?', 0, 59)
        parse(r'\d{1,6}')  # microsecond

        return Datetime._datetime(data)

    def replace(self, **kwargs):
        """
        'weekday'，1~7分别代表周一到周日

        >>> Datetime(180826).replace(weekday=1).strftime('%y%m%d周%k')
        '180820周一'
        >>> Datetime(180826).replace(weekday=3).strftime('%y%m%d周%k')
        '180822周三'
        """
        a = self
        if 'weekday' in kwargs:
            if set(kwargs.keys()) & {'year', 'month', 'day'}:
                raise ValueError('weekday参数不能混合年月日的修改使用')
            a = a - a.isoweekday() + kwargs['weekday']  # 先减去当前星期几，再加上1~7的一个目标星期
            del kwargs['weekday']
        if kwargs:
            a = a.replace(**kwargs)
        return Datetime(a)

    def __add__(self, other):
        """加减数值时，单位按天处理"""
        if isinstance(other, (int, float)):
            other = datetime.timedelta(other)
        return Datetime(super().__add__(other))

    def __sub__(self, other):
        if isinstance(other, (int, float)):
            other = datetime.timedelta(other)
        return super().__sub__(other)

    def shift(self, **kwargs):
        """shift 有点像游标卡尺，可以左右两边进行加减移位操作，加减的对象可以是年月日时分秒和星期。

        >>> a = Datetime(2018, 8, 24)
        >>> a.shift(months=-1)
        <Datetime [2018-07-24T00:00:00+08:00]>
        >>> a.shift(months=-1).format("YYYYMM")
        '201807'
        >>> a.shift(weeks=1)
        <Datetime [2018-08-31T00:00:00+08:00]>
        """
        return Datetime(super().shift(**kwargs))

    def to(self, tz):
        """to 可以将一个本地时区转换成其它任意时区，例如：

        >>> a = Datetime(2020, 6, 1, 17)
        >>> a.to("utc")
        <Datetime [2020-06-01T09:00:00+08:00]>
        >>> a.to("utc").to("local")
        <Datetime [2020-06-01T09:00:00+08:00]>
        >>> a.to("America/New_York")
        <Datetime [2020-06-01T05:00:00+08:00]>
        """
        return Datetime(super().to(tz))

    def humanize(self, other=None, locale="zh", only_distance=False, granularity="auto"):
        r""" humanize 方法是相对于当前时刻表示为“多久以前”的一种可读行字符串形式，
                默认是英文格式，指定 locale 可显示相应的语言格式。

        >>> a = Datetime().shift(hours=-6)
        >>> a.humanize()
        '6小时前'
        >>> a.humanize(locale='en')
        '6 hours ago'

        这是从 https://mp.weixin.qq.com/s/DqD_PmrspMeytloV_o54IA 摘录的Arrow的文档
            并不是该Datetime类本身的文档，但意思差不多，可以参考
        """
        return super().humanize(other, locale, only_distance, granularity)

    def strftime(self, fmt='%Y/%m/%d'):
        r"""做了简单格式拓展

        官方支持的符号： https://docs.python.org/3/library/datetime.html#strftime-strptime-behavior

        我自己扩展的功能：
        %k，周几：一、二、三、四、五、六、日

        >>> dt = Datetime(2020, 2, 28, 9, 10, 52)
        >>> dt.strftime('%y%m%d')
        '200228'
        >>> dt.strftime('%Y/%m/%d')
        '2020/02/28'
        >>> dt.strftime('%y%m%d周%k')  # 我自己最常用的格式
        '200228周五'
        """
        # 先用占位符替代中文，和我扩展的%k等标记
        fmt1 = re.sub(r'([\u4e00-\u9fa5，。；？（）【】、①-⑨]|%k)', r'{placeholder}', fmt)
        tag = super().strftime(fmt1)
        if fmt1 != fmt:
            texts = re.findall(r'([\u4e00-\u9fa5，。；？（）【】、①-⑨]|%k)', fmt)
            for i in range(len(texts)):
                tag = tag.replace('{placeholder}', texts[i], 1)
            tag = tag.replace('%k', '日一二三四五六'[self.isoweekday() % 7])
        return tag



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# @Author : 陈坤泽
# @Email  : 877362867@qq.com
# @Date   : 2021/05/26 17:24

import json
import sys
import subprocess
import os.path as osp

try:
    from qtpy import QtWidgets
except ModuleNotFoundError:
    subprocess.run(['pip3', 'install', 'QtPy'])
    from qtpy import QtWidgets

from qtpy import QtGui
from qtpy.QtWidgets import QFrame, QInputDialog, QApplication
from PyQt5.QtCore import pyqtSignal

from pyxllib.prog.newbie import CvtType

here = osp.dirname(osp.abspath(__file__))


class QHLine(QFrame):
    """ https://stackoverflow.com/questions/5671354/how-to-programmatically-make-a-horizontal-line-in-qt """

    def __init__(self):
        super(QHLine, self).__init__()
        self.setFrameShape(QFrame.HLine)
        self.setFrameShadow(QFrame.Sunken)


class QVLine(QFrame):
    def __init__(self):
        super(QVLine, self).__init__()
        self.setFrameShape(QFrame.VLine)
        self.setFrameShadow(QFrame.Sunken)


class XlLineEdit(QtWidgets.QLineEdit):
    correctChanged = pyqtSignal(object)  # 符合数值类型的修改
    wrongChanged = pyqtSignal(str)  # 不符合数值类型的修改

    def __init__(self, text=None, parent=None, *, valcvt=None):
        """
        :param valcvt: 数值类型转换器
        """
        super().__init__(str(text), parent)

        def check():
            # TODO 目前是强制重置样式，可以考虑怎么保留原样式基础上修改属性值
            s = self.text()
            try:
                if valcvt:
                    s = valcvt(s)
                self.setStyleSheet('')
                self.setToolTip('')
                self.correctChanged.emit(s)
            except ValueError:
                self.setStyleSheet('background-color: lightpink;')
                self.setToolTip(f'输入数据不是{valcvt}类型')
                self.wrongChanged.emit(s)

        self.textChanged.connect(check)
        if text:
            check()

        # self.setStyleSheet(self.styleSheet() + 'qproperty-cursorPosition: 0;')
        self.setStyleSheet(self.styleSheet())


class XlComboBox(QtWidgets.QComboBox):
    # 这个控件一般没有类型检查，但在支持填入自定义值时，是存在类型错误问题的
    #   但在工程上还是依然写了 correctChanged，方便下游任务统一接口
    correctChanged = pyqtSignal(object)  # 符合数值类型的修改
    wrongChanged = pyqtSignal(str)  # 不符合数值类型的修改

    def __init__(self, parent=None, *, text=None, items=None, valcvt=None, editable=False):
        """
        """
        # 1 基础设置
        super().__init__(parent)
        self.reset_items(items)
        self.editable = editable
        if self.editable:
            self.setEditable(True)

        # 2 检查功能
        def check(s):
            try:
                if valcvt:
                    s = valcvt(s)

                if self.editable:  # 支持自定义值
                    self.setStyleSheet('')
                    self.setToolTip('')
                    self.correctChanged.emit(s)
                elif s not in self.items_set:  # 不支持自定义值，但出现自定义值
                    self.setStyleSheet('background-color: yellow;')
                    self.setToolTip(f'不在清单里的非法值')
                    self.wrongChanged.emit(s)
                else:  # 不支持自定义值，且目前值在清单中
                    self.setEditable(False)
                    self.setStyleSheet('')
                    self.setToolTip('')
                    self.correctChanged.emit(s)
            except ValueError:
                self.setStyleSheet('background-color: lightpink;')
                self.setToolTip(f'输入数据不是{valcvt}类型')
                self.wrongChanged.emit(s)

        self.currentTextChanged.connect(check)
        # self.wrongChanged.connect(lambda s: print('非法值：', s))  # 可以监控非法值

        # 3 是否有预设值
        if text:
            self.setText(text)

        # 4 补充格式
        # self.setStyleSheet(self.styleSheet() + 'qproperty-cursorPosition: 0;')
        self.setStyleSheet(self.styleSheet())

    def setText(self, text):
        text = str(text)
        if text not in self.items and not self.editable:
            # 虽然不支持editable，但是出现了意外值，需要强制升级为可编辑
            self.setEditable(True)
        self.setCurrentText(text)

    def reset_items(self, items):
        # 1 存储配置清单
        self.clear()
        self.raw_items = items  # noqa
        self.items = [str(x) for x in items if x is not None]  # noqa
        self.items_set = set(self.items)  # noqa 便于判断是否存在的集合类型
        self.addItems(self.items)

        # 2 画出 元素值、分隔符
        cnt = 0
        for i in range(len(items)):
            if items[i] is None:
                self.insertSeparator(i - cnt)  # 不过这个分割符没那么显眼
                cnt += 1


def get_input_widget(items=None, cur_value=None, *, valcvt=None,
                     n_widget=1, enabled=True,
                     correct_changed=None):
    """ 根据items参数情况，智能判断生成对应的windet

    :param items:
        None, 普通的文本编辑框
        普通数组，下拉框  （可以用None元素表示分隔符）
            list，除了列表中枚举值，也支持自定义输入其他值
            tuple，只能用列表中的枚举值
        多级嵌套数组，多级下拉框  （未实装）  （一般都是不可改的，并且同类型的数据）
            [('福建', [('龙岩', ['连城', '长汀', ...], ...)]), ('北京', ...)]
            这种情况会返回多个widget
    :param cur_value: 当前显示的文本值
    :param valcvt: 数值类型转换函数，非法时返回ValueError  （未实装）
        很多输入框是传入文本，有时需要转为int、float、list等类型
        支持输入常见类型转换的字符串名称，比如int、float
    :param correct_changed: 文本改变时的回调函数
    :param n_widget: 配合items为嵌套数组使用，需要指定嵌套层数
        此时cur_value、cvt、enabled、text_changed等系列值可以传入n_widget长度的list
    :param enabled: 是否可编辑
    """
    if n_widget > 1:
        raise NotImplementedError

    # 1 封装类型检查功能
    if isinstance(valcvt, str):
        cvtfunc = CvtType.factory(valcvt)
    else:
        cvtfunc = valcvt

    # 2 正式生成控件
    if isinstance(items, (list, tuple)):
        # 带有 items 的字段支持候选下拉菜单
        w = XlComboBox(text=cur_value, items=items, valcvt=cvtfunc, editable=isinstance(items, list))
    elif items is None:
        # 普通填充框
        w = XlLineEdit(cur_value, valcvt=cvtfunc)
    else:
        raise ValueError(f'{type(items)}')

    # 3 通用配置
    if callable(correct_changed):
        w.correctChanged.connect(correct_changed)
    if not enabled:
        w.setEnabled(enabled)

    return w


def newIcon(icon):
    icons_dir = osp.join(here, "../icons")
    return QtGui.QIcon(osp.join(":/", icons_dir, "%s.png" % icon))


def newAction(
        parent,
        text,
        slot=None,
        shortcut=None,
        icon=None,
        tip=None,
        checkable=False,
        enabled=True,
        checked=False,
):
    """Create a new action and assign callbacks, shortcuts, etc."""
    a = QtWidgets.QAction(text, parent)
    if icon is not None:
        a.setIconText(text.replace(" ", "\n"))
        a.setIcon(newIcon(icon))
    if shortcut is not None:
        if isinstance(shortcut, (list, tuple)):
            a.setShortcuts(shortcut)
        else:
            a.setShortcut(shortcut)
    if tip is not None:
        a.setToolTip(tip)
        a.setStatusTip(tip)
    if slot is not None:
        a.triggered.connect(slot)
    if checkable:
        a.setCheckable(True)
    a.setEnabled(enabled)
    a.setChecked(checked)
    return a


class XlActionFunc:
    """ 跟action有关的函数调用封装 （该类也可以作为基础的check版本使用）

    一般逻辑结构是这样：
        有个可运行功能的函数func，运行时会配置一些需要存储起来的变量值value
        并且功能需要关联action，绑定到menu等菜单中时，可以使用该装饰器
    """

    def __init__(self, parent, title, value=None, checked=None, **kwargs):
        self.parent = parent
        self.title = title
        self.value = value
        self.checked = bool(checked)
        self.checkable = checked is not None

        self.action = newAction(self.parent, self.parent.tr(self.title), self.__call__,
                                checkable=self.checkable, checked=self.checked, **kwargs)

    def __call__(self, checked):
        self.checked = checked


class GetMultiLineTextAction(XlActionFunc):
    """ 该类value是直接存储原始的完整文本内容 """

    def __call__(self, checked):
        super().__call__(checked)
        self.value = self.value or ''
        inputs = QInputDialog.getMultiLineText(self.parent, self.title, '编辑文本：', self.value)
        if inputs[1]:  # “确定操作” 才更新属性
            self.value = inputs[0]


class GetItemsAction(XlActionFunc):
    """ 该类value目前是存储为list类型 """

    def __call__(self, checked):
        super().__call__(checked)
        self.value = self.value or []
        inputs = QInputDialog.getMultiLineText(self.parent, self.title, '编辑多行文本：',
                                               '\n'.join(self.value))
        if inputs[1]:  # “确定操作” 才更新属性
            self.value = inputs[0].splitlines()


class GetJsonAction(XlActionFunc):
    """ 该类value是直接存储原始的完整文本内容 """

    def __call__(self, checked):
        super().__call__(checked)
        self.value = self.value or ''
        inputs = QInputDialog.getMultiLineText(self.parent, self.title, '编辑json数据：',
                                               json.dumps(self.value, indent=2))
        if inputs[1]:  # “确定操作” 才更新属性
            self.value = json.loads(inputs[0])


def main_qapp(window):
    """ 执行Qt应用 """
    app = QApplication(sys.argv)
    window.show()  # 展示窗口
    sys.exit(app.exec_())



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# @Author : 陈坤泽
# @Email  : 877362867@qq.com
# @Date   : 2021/01/13 15:25

import os
import shutil


def unpack_zipfile(filename, extract_dir):
    """ 为了修复zipfile底层的中文解压乱码问题，修改了shutil._UNPACK_FORMATS的底层功能
    """
    from pyxllib.stdlib import zipfile
    from pyxllib.xl import Dir, File
    from shutil import ReadError

    zip = zipfile.ZipFile(filename)
    if not zipfile.is_zipfile(filename):
        raise ReadError("%s is not a zip file" % filename)
    try:
        for info in zip.infolist():
            name = info.filename

            # don't extract absolute paths or ones with .. in them
            if name.startswith('/') or '..' in name:
                continue

            target = os.path.join(extract_dir, *name.split('/'))
            if not target:
                continue

            Dir(File(target).parent).ensure_dir()
            if not name.endswith('/'):
                # file
                data = zip.read(info.filename)
                f = open(target, 'wb')
                try:
                    f.write(data)
                finally:
                    f.close()
                    del data
    finally:
        zip.close()


# 解决unzip中文乱码问题
shutil._UNPACK_FORMATS['zip'] = (['.zip'], unpack_zipfile, [], "ZIP file")



from pyxllib.xl import distribute_package

distribute_package('D:/slns/pyxllib', '0.1.21')

# 注意这个版本机制不太一样，ui界面标题的版本是另设的
# distribute_package('D:/slns/xllabelme', '1.8')

# distribute_package('D:/slns/xlcocotools', '0.0.3')  # 还未上传，待上传

# distribute_package('D:/slns/pyxlpr', '0.0.45', 'xlpr')



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# @Author : 陈坤泽
# @Email  : 877362867@qq.com
# @Date   : 2020/06/03 09:52

import math
import subprocess

try:
    from bidict import bidict
except ModuleNotFoundError:
    subprocess.run(['pip3', 'install', 'bidict'])
    from bidict import bidict

try:
    import sqlalchemy
except ModuleNotFoundError:
    subprocess.run(['pip3', 'install', 'sqlalchemy'])
    subprocess.run(['pip3', 'install', 'mysqlclient'])
    import sqlalchemy

import pandas as pd

from pyxllib.file.specialist import File

SQL_LIB_ACCOUNT_FILE = File(__file__).parent / 'sqllibaccount.pkl'


def create_account_df(file='sqllibaccount.pkl'):
    """请在这里设置您个人的账户密码，并在运行完后，销毁明文信息"""
    df = pd.DataFrame.from_records([
        ['ckz', 'rm.sbsql.rds.aliyuncs.com', '', '', 'dddddd'],
        ['ckzlocal', '0.0.0.0', '', '', 'eeeeee'],
    ], columns=['index_name', 'host', 'port', 'user', 'passwd'])
    df['port'] = df['port'].replace('', '3306')  # 没写端口的默认值
    df['user'] = df['user'].replace('', 'root')  # 没写用户名的默认值
    df['passwd'] = df['passwd'].replace('', '123456')  # 没写密码的默认值
    df.set_index('index_name', inplace=True)
    File(file).write(df)


class SqlEngine:
    """mysql 通用基础类
    """

    def __init__(self, alias=None, database=None, *,
                 user='root', passwd='123456', host=None, port='3306',
                 connect_timeout=None, account_file_path=None):
        """ 初始化需要连接数据库

        :param alias: 数据库的简化别名，为了方便快速调用及隐藏明文密码
            使用该参数将会覆盖掉已有的user、passwd、host、port参数值
            例如我自己设置的别名有：
                ckz，我自己阿里云上的个人数据库
                ckzlocal，本PC开的数据库
        :param account_file_path: 使用alias时才有效
            该参数指定存储账号信息的pkl文件所在位置，注意pkl的格式必须用类似下述的代码方式生成
            默认从与该脚本同目录下的 sqllibaccount.pkl 文件获取

        :param database: 数据库名称
            例如在快乐做教研时一些相关数据库名：
                tr，教研
                tr_develop，教研开发数据
                tr_test，教研测试数据

        :param connect_timeout: 连接超时时等待秒数
            如果设置，建议2秒以上

        :return:
        """

        # 1 读取地址、账号信息
        if alias:
            if account_file_path is None:
                account_file_path = File(SQL_LIB_ACCOUNT_FILE)
            # dprint(alias,account_file_path)
            record = File(account_file_path).read().loc[alias]  # 从文件读取账号信息
            user, passwd, host, port = record.user, record.passwd, record.host, record.port

        # 2 '数据库类型+数据库驱动名称://用户名:口令@机器地址:端口号/数据库名'
        address = f'mysql+mysqldb://{user}:{passwd}@{host}:{port}/{database}?charset=utf8mb4'
        # 3 存储成员
        self.alias, self.database = alias, database
        connect_args = {"connect_timeout": connect_timeout} if connect_timeout else {}
        self.engine = sqlalchemy.create_engine(address, connect_args=connect_args)

    def query(self, sql, index_col=None, coerce_float=True, params=None,
              parse_dates=None, columns=None, chunksize=None):
        """本质上就是pd.read_sql函数

        pd.read_sql()知道这些就够用了 - 漫步量化 - CSDN博客:
        https://blog.csdn.net/The_Time_Runner/article/details/86601988

        官方文档：pandas.read_sql — pandas 0.25.1 documentation:
        https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.read_sql.html

        :param sql: 命令字符串，可以写%s设置
            这里我做了扩展：
                单语句: '...'
                多语句: ['...', '...']，默认\n隔开
        :param index_col: 设为索引列
        :param coerce_float: 设为float类型的列
        :param params: 和sql命令相关，具体语法规则与所用的引擎相关，例如我这里是用sqlalchemy，支持用法
            'SELECT point_name FROM tr_point LIMIT %s', params=(5,))   # list、tuple等列举每个%s的值
            'SELECT point_name FROM tr_point LIMIT %(n)s', params={'n', 5})  # 是用dict关联命名参数
        :param parse_dates: 转为datetime类型的列
        :param columns: 要选取的列。一般没啥用，因为在sql命令里面一般就指定要选择的列了
        :param chunksize: 如果提供了一个整数值，那么就会返回一个generator，每次输出的行数就是提供的值的大小
        :return: DataFrame类型的数据
        """
        # 1 合并sql命令
        if isinstance(sql, str): sql = [sql]
        sql = '\n'.join(sql)

        # 含有 % 的特殊符号要转义
        # import sqlalchemy
        # sql = sqlalchemy.text(sql)

        # 2 解析结果
        res = pd.read_sql(sql, self.engine, index_col=index_col, coerce_float=coerce_float, params=params,
                          parse_dates=parse_dates, columns=columns, chunksize=chunksize)
        return res

    def execute(self, statement, *multiparams, **params):
        """本质上就是sqlalchemy.con.execute的封装

        可以这样使用：
        hsql.execute('UPDATE spell_check SET count=:count WHERE old=:old AND new=:new',
                      count=count[0] + add, old=old, new=new)
        """
        # 1 解析sql命令
        if isinstance(statement, str): statement = [statement]
        statement = '\n'.join(statement)
        statement = sqlalchemy.text(statement)

        # 2 如果设置了getdf参数
        res = self.engine.execute(statement, *multiparams, **params)
        return res

    def insert_from_df(self, df, table_name, patch_size=100, if_exists='append'):
        """将df写入con数据库的table_name表格

        190731周三18:51，TODO
        可以先用：df.to_sql('formula_stat', HistudySQL('dev', 'tr_develop').con, if_exists='replace')
        191017周四10:21，目前这函数改来改去，都还没严格测试呢~~

        这个函数开发要参考：DataFrame.to_sql()
            是因为其con参数好像不支持pymysql

        :param df: DataFrane类型表格数据
        :param table_name: 要写入的表格名
        :param patch_size: 每轮要写入的数量
            如果df很大，是无法一次性用sql语句写入的，一般要分批写
            patch_size是设置每批导入多少条数据
        :param if_exists: {'fail', 'replace', 'append'}, default 'append'
                How to behave if the table already exists.

                * fail: Raise a ValueError.
                * replace: Drop the table before inserting new values.
                * append: Insert new values to the existing table.
        """
        con = self.engine
        # TODO 增加表格是否存在的判断；我这个函数本质上只能往已存在的表格插入数据
        if if_exists == 'append':
            pass
        elif if_exists == 'replace':
            con.query(f'TRUNCATE TABLE {table_name}')
        elif if_exists == 'fail':
            raise ValueError('表格已存在')
        else:
            raise NotImplementedError

        # 1 删除table中不支持的df的列
        cols = pd.read_sql(f'SHOW COLUMNS FROM {table_name}', con)['Field']
        cols = list(set(df.columns) & set(cols))
        df = df[cols]

        # 2 将df每一行数据转成mysql语句文本
        data = []  # data[i]是第i条数据的sql文本

        # 除了nan，bool值、None值都能正常转换
        def func(x):
            # s = con.escape(str(x))
            s = x
            if s == 'nan': s = 'NULL'  # nan转为NULL
            return s

        for idx, row in df.iterrows():
            t = ', '.join(map(func, row))
            data.append('(' + t + ')')

        # 3 分批导入
        columns = '( ' + ', '.join(cols) + ' )'
        for j in range(0, math.ceil(len(data) / patch_size)):
            subdata = ',\n'.join(data[j * patch_size:(j + 1) * patch_size])
            con.execute("INSERT IGNORE INTO :a :b VALUES :c",
                        a=table_name, b=columns, c=subdata)
        con.commit()  # 更新后才会起作用


class SqlCodeGenerator:
    @staticmethod
    def keys_count(table, keys):
        codes = [f'-- 分析{table}表中，{keys}出现的种类和次数，按照出现次数从多到少排序',
                 f'SELECT {keys}, COUNT(*) cnt FROM {table} GROUP BY {keys} ORDER BY cnt DESC']
        return '\n'.join(codes)

    @staticmethod
    def one2many(table, keys, vars):
        codes = [f'-- 分析{table}表中，{keys}构成的键，对应{vars}构成的值，是否有一对多的关系，按多到少排序',
                 f'SELECT {keys}, COUNT(DISTINCT {vars}) cnt',
                 f'FROM {table} GROUP BY {keys}',
                 'HAVING cnt > 1 ORDER BY cnt DESC']
        return '\n'.join(codes)


def demo_sqlengine():
    db = SqlEngine('ckz', 'runoob')
    df = db.query('SELECT * FROM apps')
    print(df)


class MultiEnumTable:
    """多份枚举表的双向映射
        目前是用来做数据库表中的枚举值映射，但实际可以通用于很多地方

    >>> met = MultiEnumTable()
    >>> met.add_enum_table('subject', [5, 8, 6], ['语文', '数学', '英语'])
    >>> met.add_enum_table_from_dict('grade', {1: '小学', 2: '初中', 3: '高中'})

    >>> met['subject'][6]
    '英语'
    >>> met['subject'].inverse['英语']
    6

    >>> met.decode('subject', 5)
    '语文'
    >>> met.encode('subject', '数学')
    8

    >>> met.decodes('grade', [1, 3, 3, 2, 1])
    ['小学', '高中', '高中', '初中', '小学']
    >>> met.encodes('grade', ['小学', '高中', '大学', '初中', '小学'])
    [1, 3, None, 2, 1]
    """

    def __init__(self):
        self.enum_tables = dict()

    def __getitem__(self, table):
        return self.enum_tables[table]

    def add_enum_table(self, table, ids, values):
        """增加一个映射表"""
        self.enum_tables[table] = bidict({k: v for k, v in zip(ids, values)})

    def add_enum_table_from_dict(self, table, d):
        self.enum_tables[table] = bidict({k: v for k, v in d.items()})

    def set_alias(self, table, alias):
        """已有table的其他alias别名
        :param alias: list
        """
        for a in alias:
            self.enum_tables[a] = self.enum_tables[table]

    def decode(self, table, id_, default=None):
        """转明文"""
        return self.enum_tables[table].get(id_, default)

    def encode(self, table, value, default=None):
        """转id"""
        return self.enum_tables[table].inverse.get(value, default)

    def decodes(self, table, ids, default=None):
        d = self.enum_tables[table]
        return [d.get(k, default) for k in ids]

    def encodes(self, table, values, default=None):
        d = self.enum_tables[table].inverse
        return [d.get(v, default) for v in values]


def adjust_repeat_data(li, suffix='+'):
    """ 分析序列li里的值，对出现重复的值进行特殊标记去重
    :param li: list，每个元素值一般是str
    :param suffix: 通过增加什么后缀来去重
    :return: 新的无重复数值的li

    >>> adjust_repeat_data(['a', 'b', 'a', 'c'])
    ['a', 'b', 'a+', 'c']
    """
    res = []
    values = set()
    for x in li:
        while x in values:
            x += suffix
            # print(x)
        res.append(x)
        values.add(x)

    return res



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# @Author : 陈坤泽
# @Email  : 877362867@qq.com
# @Date   : 2021/04/04 17:03

""" 专门给utools的快捷命令扩展的一系列python工具库
"""

import os
import pathlib
import pyperclip
import re

import ujson
from humanfriendly import format_timespan
import pandas as pd

from pyxllib.debug.specialist import browser, TicToc, Datetime
from pyxllib.file.specialist import File, Dir


def _print_df_result(df, outfmt='text'):
    # subinput可以强制重置输出类型
    # TODO argparser可以对str类型进行参数解析？
    # if 'browser' in kwargs['subinput']:
    #     outfmt = 'browser'
    # elif 'html' in kwargs['subinput']:
    #     outfmt = 'html'
    # elif 'text' in kwargs['subinput']:
    #     outfmt = 'text'

    if outfmt == 'html':
        # content = df.to_html().replace('–', '-')  # 很多app会出现一个utf8特殊的-，转gbk编码会出错+
        try:  # utools里按gbk会有很多编码问题，如果不能直接显示，就打开浏览器看
            print(df.to_html())
        except UnicodeEncodeError:
            browser(df)
    elif outfmt == 'browser':
        browser(df)
    else:
        with pd.option_context('display.max_colwidth', -1, 'display.max_columns', 20,
                               'display.width', 200):  # 上下文控制格式
            print(df)


class UtoolsBase:
    def __init__(self, cmds, *, outfmt='text'):
        """
        :param cmds: "快捷命令"响应的命令集，即调用出可传入的输入文件、所在窗口、选中文件等信息
        :param outfmt: 输出格式
            text，纯文本
            html，html格式
            browser，用浏览器打开
        """
        # 1 删除没有值的键
        self.cmds = dict()
        for k, v in cmds.items():
            if cmds[k] not in ('{{' + k + '}}', ''):
                self.cmds[k] = v

        # 2 「快捷命令」的ClipText宏展开，如果存在连续3个单引号，会跟py字符串定义冲突，产生bug
        #   所以这段不建议用宏展开，这里该类会自动调用pyperclip获取
        #   这个可以作为可选功能关闭，但实验证明这个对性能其实没有太大影响
        if 'ClipText' not in self.cmds:
            self.cmds['ClipText'] = pyperclip.paste().replace('\r\n', '\n')

        # 3 解析json数据
        # 解析window模式的WindowInfo
        if 'WindowInfo' in self.cmds:
            self.cmds['WindowInfo'] = ujson.loads(self.cmds['WindowInfo'].encode('utf8'))

        # 解析files模式的MatchedFiles
        if 'MatchedFiles' in self.cmds:
            # 右边引用没写错。因为MatchedFiles本身获取内容不稳定，可能会有bug，取更稳定的payload来初始化MatchedFiles
            self.cmds['MatchedFiles'] = ujson.loads(self.cmds['payload'].encode('utf8'))
            # 在有些软件中，比如everything，是可以选中多个不同目录的文件的。。。所以严格来说不能直接按第一个文件算pwd
            # 但这个讯息是可以写一个工具分析的
            # self.cmds['pwd'] = os.path.dirname(self.cmds['MatchedFiles'][0]['path'])

            # TODO 如果选中的文件全部是同一个目录下的，要存储pwd
            dirs = {os.path.dirname(f['path']) for f in self.cmds['MatchedFiles']}
            if len(dirs) == 1:
                self.cmds['pwd'] = list(dirs)[0]

        # 4 当前工作目录
        # 要根据场景切换pwd
        #   files默认工作目录是app: C:\Users\chen\AppData\Local\Programs\utools
        #   window默认工作目录是exe
        # window模式存储的pwd跟各种软件有关
        #   explorer就是当前窗口，没问题
        #   onenote存的是桌面
        if 'pwd' in self.cmds:
            os.chdir(self.cmds['pwd'])

        # 5 subinput可以用字典结构扩展参数
        if 'subinput' in self.cmds:
            if re.match(r'\{.+\}$', self.cmds['subinput']):
                ext_kwargs = eval(self.cmds['subinput'])
                self.cmds.update(ext_kwargs)

        self.outfmt = self.cmds['outfmt'] if 'outfmt' in self.cmds else outfmt

    def check_cmds(self):
        """ 显示所有参数值 """
        df = pd.DataFrame.from_records([(k, v) for k, v in self.cmds.items()],
                                       columns=['key', 'content'])
        _print_df_result(df, self.outfmt)


class UtoolsRegex(UtoolsBase):
    def __init__(self, cmds, *, outfmt='text'):
        super().__init__(cmds, outfmt=outfmt)

    def coderegex(self):
        tt = TicToc()
        text = self.cmds['ClipText']
        eval(self.cmds['subinput'])
        print(f'finished in {format_timespan(tt.tocvalue())}.')


class UtoolsFile(UtoolsBase):
    """ 文件相关操作工具 """

    def __init__(self, cmds, *, outfmt='text'):
        super().__init__(cmds, outfmt=outfmt)

        # 如果是window等模式，补充 MatchedFiles
        if 'MatchedFiles' in self.cmds:
            self.paths = [pathlib.Path(f['path']) for f in self.cmds['MatchedFiles']]
        else:
            self.paths = [pathlib.Path(os.path.abspath(f)).resolve() for f in os.listdir('.')]

    @classmethod
    def is_image(cls, f):
        return bool(re.match('jpe?g|png|gif|bmp', f.suffix.lower()[1:]))

    def codefile(self):
        """ 多用途文件处理器

        核心理念是不设定具体功能，而是让用户在 subinput 自己写需要执行的py代码功能
        而在subinput，可以用一些特殊的标识符来传参

        以file为例，有4类参数：
            file，处理当前目录 文件
            rfile，递归处理所有目录下 文件
            files，当前目录 所有文件
            rfiles，递归获取所有目录下 所有文件

        类型上，file可以改为
            dir，只分析目录
            path，所有文件和目录

        扩展了一些特殊的类型：
            imfile，图片文件
        """
        from functools import reduce
        try:
            from xlproject.kzconfig import KzDataSync
            from pyxllib.data.label import reduce_labelme_jsonfile
        except ModuleNotFoundError:
            pass

        tt = TicToc()

        # 1 获得所有标识符
        # 每个单词前后都有空格，方便定界
        keywords = filter(lambda x: re.match(r'[a-zA-Z_]+$', x),
                          set(re.findall(r'\.?[a-zA-Z_]+\(?', self.cmds['subinput'])))
        keywords = ' ' + ' '.join(keywords) + ' '
        # print('keywords:', keywords)

        # 2 生成一些备用的智能参数
        # 一级目录下选中的文件
        paths = self.paths
        # 用正则判断一些智能参数是否要计算，注意不能只判断files，如果出现file，也是要引用files的
        files = [File(p) for p in paths if p.is_file()] if re.search(r'files? ', keywords) else []
        imfiles = [f for f in files if self.is_image(f)] if re.search(r' imfiles? ', keywords) else []
        # 出现r系列的递归操作，都是要计算出dirs的
        dirs = [Dir(p) for p in paths if p.is_dir()] if re.search(r' (dirs?|r[a-zA-Z_]+) ', keywords) else []

        # 递归所有的文件
        rpaths = reduce(lambda x, y: x + y.select('**/*').subpaths(), [paths] + dirs) \
            if re.search(r' (r[a-zA-Z_]+) ', keywords) else []
        rfiles = [File(p) for p in rpaths if p.is_file()] if re.search(r' (r(im)?files?) ', keywords) else []
        rimfiles = [f for f in rfiles if self.is_image(f)] if re.search(r' rimfiles? ', keywords) else []
        rdirs = [Dir(p) for p in rpaths if p.is_dir()] if re.search(r' (rdirs?) ', keywords) else []

        # 3 判断是否要智能开循环处理
        m = re.search(r' (r?(path|(im)?file|dir)) ', keywords)
        if m:
            name = m.group(1)
            objs = eval(name + 's')
            print('len(' + name + 's)=', len(objs))
            for x in objs:
                locals()[name] = x
                eval(self.cmds['subinput'])
        else:  # 没有的话就直接处理所有文件
            eval(self.cmds['subinput'])

        # 4 运行结束标志
        print(f'finished in {format_timespan(tt.tocvalue())}.')


def clipboard_paste(func):
    def wrapper(*args, **kwargs):
        import pyautogui
        s = func(*args, **kwargs)

        # print(s, end='')
        pyperclip.copy(s)
        # pyperclip.paste()  # 这个没用
        # pyautogui.write(s)  # 这个也不能写入中文
        pyautogui.hotkey('ctrl', 'v')  # 目前来看就这个方法最靠谱

        return s

    return wrapper


class UtoolsText(UtoolsBase):
    """ 目录路径生成工具 """

    @clipboard_paste
    def wpath(self):
        """ windows上的路径 """
        from xlproject.kzpaths import CommonDir

        p = getattr(CommonDir, self.cmds['subinput'])
        # slns比较特别，本地要重定向到D:/slns
        p = str(p).replace('D:/home/chenkunze/slns', 'D:/slns')
        p = p.replace('/', '\\')
        return p

    @clipboard_paste
    def upath(self):
        """ unix上的路径 """
        from xlproject.kzpaths import CommonDir

        p = getattr(CommonDir, self.cmds['subinput'])
        # 因为用了符号链接，实际位置会变回D:/slns，这里需要反向替换下
        p = str(p).replace('D:/slns', 'D:/home/chenkunze/slns')
        p = str(p)[2:]
        return p

    def wdate(self):
        """ 输入一周的简略日期值 """
        import pyautogui

        start = int(self.cmds['subinput'])

        for i in range(7):
            dt = Datetime(start) + i
            pyperclip.copy(dt.strftime('%y%m%d周%k'))
            # pyperclip.paste()  # 这个没用
            # pyautogui.write('210503周一')  # 这个也没用
            pyautogui.hotkey('ctrl', 'v')
            pyautogui.press('down')


if __name__ == '__main__':
    with TicToc('utools'):
        pass



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# @Author : 陈坤泽
# @Email  : 877362867@qq.com
# @Date   : 2021/06/06 16:57

from collections import Counter
import subprocess
import re

try:
    import ahocorasick
except ModuleNotFoundError:
    subprocess.run(['pip3', 'install', 'pyahocorasick'])
    import ahocorasick


def make_automaton(words):
    """ 根据输入的一串words模式，生成一个AC自动机 """
    a = ahocorasick.Automaton()
    for index, word in enumerate(words):
        a.add_word(word, (index, word))
    a.make_automaton()
    return a


def count_words(content, word, scope=2, exclude=None):
    # 1 统计所有词汇出现次数
    c = Counter()
    c += Counter(re.findall(f'.{{,{scope}}}{word}.{{,{scope}}}', content))
    # 2 排除掉不处理的词 （注意因为这里每句话都已经是被筛选过的，所以处理比较简单，并不需要复杂到用区间集处理）
    if exclude:
        new_c = Counter()
        a = make_automaton(exclude)  # 创建AC自动机
        for k in c.keys():
            if not next(a.iter(k), None):
                # 如果k没匹配到需要排除的词汇，则拷贝到新的计数器
                new_c[k] = c[k]
        c = new_c
    return c



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# @Author : 陈坤泽
# @Email  : 877362867@qq.com
# @Date   : 2020/06/02 20:16

"""
xml等网页结构方面的处理
"""
import collections
from collections import Counter, defaultdict
import re
import textwrap

import requests
import pandas as pd
import bs4
from bs4 import BeautifulSoup

from pyxllib.prog.newbie import typename
from pyxllib.debug.pupil import dprint
from pyxllib.text.newbie import xldictstr
from pyxllib.text.pupil import listalign, int2myalphaenum, shorten, ensure_gbk
from pyxllib.file.specialist import File, Dir

____section_1_dfs_base = """
一个通用的递归功能
"""


def dfs_base(node, *,
             child_generator=None, select_depth=None, linenum=False,
             mystr=None, msghead=True, lsstr=None, show_node_type=False, prefix='    '):
    """ 输入一个节点node，以及该节点当前depth

    :param prefix: 缩进格式，默认用4个空格
    :param node: 节点
    :param child_generator: 子节点生成函数
        函数支持输入一个节点参数
        返回一个子节点列表
    :param select_depth: 要显示的深度
        单个数字：获得指定层
        Sequences： 两个整数，取出这个闭区间内的层级内容
    :param mystr: 自定义单个节点字符串方式
        标准是输入2个参数 mystr(node, depth)，返回字符串化的结果，记得前缀缩进也要自己控制的！
        也可以只输入一个参数 mystr(node)：
            这种情况会自动按照每层4个空格进行缩进
    :param lsstr: 自定义整个列表的字符串化方法，在mystr的基础上调控更加灵活，但要写的代码也更多
    :param linenum：节点从1开始编号
        行号后面，默认会跟一个类似Excel列名的字母，表示层级深度
    :param msghead: 第1行输出一些统计信息
    :param show_node_type:
    :return 返回一个遍历清单ls
        ls的每个元素是一个列表
            第1个值是depth
            第2个值是节点ref

    Requires
        textwrap：用到shorten
        align.listalign：生成列编号时对齐
    """

    # 1 子节点生成器，与配置
    def bs4_child_generator(node):
        try:
            return node.children
        except AttributeError:
            return []

    # 配置子节点生成器
    if not child_generator:
        child_generator = bs4_child_generator

    # 2 dfs实际实现代码，获得节点清单
    def inner(node, depth=0):
        """dfs实际实现代码
        TODO：把depth过滤写进inner不生成？！ 不过目前还是按照生成整棵树处理，能统计到一些信息。
        """
        ls = [[node, depth]]
        for t in child_generator(node):
            ls += inner(t, depth + 1)
        return ls

    ls = inner(node)
    total_node = len(ls)
    total_depth = max(map(lambda x: x[1], ls))
    head = f'总节点数：1~{total_node}，总深度：0~{total_depth}'

    # 4 过滤与重新整理ls（select_depth）
    logo = True
    cnt = 0
    tree_num = 0
    if isinstance(select_depth, int):

        for i in range(total_node):
            if ls[i][1] == select_depth:
                ls[i][1] = 0
                cnt += 1
                logo = True
            elif ls[i][1] < select_depth and logo:  # 遇到第1个父节点添加一个空行
                ls[i] = ''
                tree_num += 1
                logo = False
            else:  # 删除该节点，不做任何显示
                ls[i] = None
        head += f'；挑选出的节点数：{cnt}，所选深度：{select_depth}，树数量：{tree_num}'

    elif hasattr(select_depth, '__getitem__'):
        for i in range(total_node):
            if select_depth[0] <= ls[i][1] <= select_depth[1]:
                ls[i][1] -= select_depth[0]
                cnt += 1
                logo = True
            elif ls[i][1] < select_depth[0] and logo:  # 遇到第1个父节点添加一个空行
                ls[i] = ''
                tree_num += 1
                logo = False
            else:  # 删除该节点，不做任何显示
                ls[i] = None
        head += f'；挑选出的节点数：{cnt}，所选深度：{select_depth[0]}~{select_depth[1]}，树数量：{tree_num}'
    """注意此时ls[i]的状态，有3种类型
        (node, depth)：tuple类型，第0个元素是node对象，第1个元素是该元素所处层级
        None：已删除元素，但为了后续编号方便，没有真正的移出，而是用None作为标记
        ''：已删除元素，但这里涉及父节点的删除，建议此处留一个空行
    """

    # 5 格式处理
    def default_mystr(node, depth):
        s1 = prefix * depth
        s2 = typename(node) + '，' if show_node_type else ''
        s3 = textwrap.shorten(str(node), 200)
        return s1 + s2 + s3

    def default_lsstr(ls):
        nonlocal mystr
        if not mystr:
            mystr = default_mystr
        else:
            try:  # 测试两个参数情况下是否可以正常运行
                mystr('', 0)
            except TypeError:
                # 如果不能正常运行，则进行封装从而支持2个参数
                func = mystr

                def str_plus(node, depth):  # 注意这里函数名要换一个新的func
                    return prefix * depth + func(node)

                mystr = str_plus

        line_num = listalign(range(1, total_node + 1))
        res = []
        for i in range(total_node):
            if ls[i] is not None:
                if isinstance(ls[i], str):  # 已经指定该行要显示什么
                    res.append(ls[i])
                else:
                    if linenum:  # 增加了一个能显示层级的int2excel_col_name
                        res.append(line_num[i] + int2myalphaenum(ls[i][1]) + ' ' + mystr(ls[i][0], ls[i][1]))
                    else:
                        res.append(mystr(ls[i][0], ls[i][1]))

        s = '\n'.join(res)
        return s

    if not lsstr:
        lsstr = default_lsstr

    s = lsstr(ls)

    # 是否要添加信息头
    if msghead:
        s = head + '\n' + s

    return s


def treetable(childreds, parents, arg3=None, nodename_colname=None):
    """输入childres子结点id列表，和parents父结点id列表
    两个列表长度必须相等
    文档：http://note.youdao.com/noteshare?id=126200f45d301fcb4364d06a0cae8376

    有两种调用形式
    >> treetable(childreds, parents)  --> DataFrame  （新建df）
    >> treetable(df, child_colname, parent_colname)  --> DataFrame （修改后的df）

    返回一个二维列表
        新的childreds （末尾可能回加虚结点）
        新的parents
        函数会计算每一行childred对应的树排序后的排序编号order
        以及每个节点深度depth

    >> ls1 = [6, 2, 4, 5, 3], ls2 = [7, 1, 2, 2, 1], treetable(ls1, ls2)
          child_id   parent_id   depth     tree_order    tree_struct
        5        7     root        1           1         = = 7
        0        6        7        2           2         = = = = 6
        6        1     root        1           3         = = 1
        1        2        1        2           4         = = = = 2
        2        4        2        3           5         = = = = = = 4
        3        5        2        3           6         = = = = = = 5
        4        3        1        2           7         = = = = 3
    """
    # 0 参数预处理
    if isinstance(childreds, pd.DataFrame):
        df = childreds
        child_colname = parents
        parent_colname = arg3
        if not arg3: raise TypeError
        childreds = df[child_colname].tolist()
        parents = df[parent_colname].tolist()
    else:
        df = None

    # 1 建立root根节点，确保除了root其他结点都存在记录
    lefts = set(parents) - set(childreds)  # parents列中没有在childreds出现的结点
    cs, ps = list(childreds), list(parents)

    if len(lefts) == 0:
        # b_left为空一定有环，b_left不为空也不一定是正常的树
        raise ValueError('有环，不是树结构')
    elif len(lefts) == 1:  # 只有一个未出现的结点，那么它既是根节点
        root = list(lefts)[0]
    else:  # 多个父结点没有记录，则对这些父结点统一加一个root父结点
        root = 'root'
        allnode = set(parents) | set(childreds)  # 所有结点集合
        while root in allnode: root += '-'  # 一直在末尾加'-'，直到这个结点是输入里未出现的
        # 添加结点
        lefts = list(lefts)
        lefts.sort(key=lambda x: parents.index(x))
        for t in lefts:
            cs.append(t)
            ps.append(root)

    n = len(cs)
    depth, tree_order, len_childs = [-1] * n, [-1] * n, [0] * n

    # 2 构造父结点-孩子结点的字典dd
    dd = defaultdict(list)
    for i in range(n): dd[ps[i]] += [i]

    # 3 dfs
    cnt = 1

    def dfs(node, d):
        """找node的所有子结点"""
        nonlocal cnt
        for i in dd.get(node, []):
            tree_order[i], depth[i], len_childs[i] = cnt, d, len(dd[cs[i]])
            cnt += 1
            dfs(cs[i], d + 1)

    dfs(root, 1)

    # 4 输出格式
    tree_struct = list(map(lambda i: f"{'_ _ ' * depth[i]}{cs[i]}" + (f'[{len_childs[i]}]' if len_childs[i] else ''),
                           range(n)))

    if df is None:
        ls = list(zip(cs, ps, depth, tree_order, len_childs, tree_struct))
        df = pd.DataFrame.from_records(ls, columns=('child_id', 'parent_id',
                                                    'depth', 'tree_order', 'len_childs', 'tree_struct'))
    else:
        k = len(df)
        df = df.append(pd.DataFrame({child_colname: cs[k:], parent_colname: ps[k:]}), sort=False, ignore_index=True)
        if nodename_colname:
            tree_struct = list(
                map(lambda i: f"{'_ _ ' * depth[i]}{cs[i]} {df.iloc[i][nodename_colname]}"
                              + (f'[{len_childs[i]}]' if len_childs[i] else ''), range(n)))
        df['depth'], df['tree_order'], df['len_childs'], df['tree_struct'] = depth, tree_order, len_childs, tree_struct
    df.sort_values('tree_order', inplace=True)  # 注意有时候可能不能排序，要维持输入时候的顺序
    return df


def treetable_flatten(df, *, reverse=False, childid_colname='id', parentid_colname='parent_id', format_colname=None):
    """获得知识树横向展开表：列为depth-3, depth-2, depth-1，表示倒数第3级、倒数第2级、倒数第1级
    :param df: DataFrame数据
    :param reverse:
        False，正常地罗列depth1、depth2、depth3...等结点信息
        True，反向列举所属层级，即显示倒数第1层parent1，然后是倒数第2层parent2...
    :param childid_colname: 孩子结点列
    :param parentid_colname: 父结点列
    :param format_colname: 显示的数值
        None，默认采用 childid_colname 的值
        str，某一列的名称，采用那一列的值（可以实现设置好格式）
    :return:
    """
    # 1 构造辅助数组
    if format_colname is None: format_colname = parentid_colname
    parentid = dict()  # parentid[k] = v， 存储结点k对应的父结点v
    nodeval = dict()  # nodeval[k] = v，  存储结点k需要显示的数值情况
    if len(df[df.index.duplicated()]):
        dprint(len(set(df.index)), len(df.index))  # 有重复index
        raise ValueError

    for idx, row in df.iterrows():
        parentid[row[childid_colname]] = row[parentid_colname]
        nodeval[row[childid_colname]] = str(row[format_colname])

    # 2 每个结点往上遍历出所有父结点
    parents = []
    for idx, row in df.iterrows():
        ps = [nodeval[row[childid_colname]]]  # 包含结点自身的所有父结点名称
        p = row[parentid_colname]
        while p in parentid:
            ps.append(nodeval[p])
            p = parentid[p]
        parents.append(ps)
    num_depth = max(map(len, parents), default=0)

    # 3 这里可以灵活调整最终要显示的格式效果
    df['parents'] = parents
    if reverse:
        for j in range(num_depth, 0, -1): df[f'depth-{j}'] = ''
        for idx, row in df.iterrows():
            for j in range(1, len(row.parents) + 1):
                df.loc[idx, f'depth-{j}'] = row.parents[j - 1]
    else:
        for j in range(num_depth): df[f'depth{j}'] = ''
        for idx, row in df.iterrows():
            for j in range(len(row.parents)):
                df.loc[idx, f'depth{j}'] = row.parents[-j - 1]
    df.drop('parents', axis=1, inplace=True)
    return df


____section_2_xml = """
xml相关的一些功能函数
"""


def readurl(url):
    """从url读取文本"""
    r = requests.get(url)
    soup = BeautifulSoup(r.text, 'lxml')
    s = soup.get_text()
    return s


____section_3_xmlparser = """
"""


def tag_name(t):
    """输入一个bs4的Tag或NavigableString，
    返回tag.name或者'NavigableString'
    """
    if t.name:
        return t.name
    elif isinstance(t, bs4.NavigableString):
        return 'NavigableString'
    else:
        dprint(t)  # 获取结点t名称失败
        return None


def subtag_names(t):
    """列出结点t的所有直接子结点（花括号后面跟的数字是连续出现次数）
    例如body的： p{137}，tbl，p{94}，tbl，p{1640}，sectPr
    """

    def counter(m):
        s1 = m.group(1)
        n = (m.end(0) - m.start(0)) // len(s1)
        s = s1[:-1] + '{' + str(n) + '}'
        if m.string[m.end(0) - 1] == '，':
            s += '，'
        return s

    if t.name and t.contents:
        s = '，'.join(map(tag_name, t.contents)) + '，'
        s = re.sub(r'([^，]+，)(\1)+', counter, s)
    else:
        s = ''
    if s and s[-1] == '，':
        s = s[:-1]
    return s


class XmlParser:
    def __init__(self, node=None):
        """两种初始化方式
            提供node：用某个bs4的PageElement等对象初始化
            未提供node，一般是方便给MyBs4等类继承使用
        """
        if node:  # TODO：可以扩展，支持不同类型的初始化
            self._node = node

    def node(self):
        """获得xml结点的接口函数"""
        return self._node if getattr(self, '_node') else self

    def treestruct_raw(self, **kwargs):
        """查看树形结构的raw版本
        各参数含义详见dfs_base
        """
        # 1 先用dfs获得基本结果
        s = dfs_base(self.node(), **kwargs)
        return s

    def treestruct_brief(self, linenum=True, prefix='- ', **kwargs):
        """查看树形结构的简洁版
        """

        def mystr(node):
            # if isinstance(node, (bs4.ProcessingInstruction, code4101py.stdlib.bs4.ProcessingInstruction)):
            if isinstance(node, bs4.ProcessingInstruction):
                s = 'ProcessingInstruction，' + str(node)
            # elif isinstance(node, (bs4.Tag, code4101py.stdlib.bs4.Tag)):
            elif isinstance(node, bs4.Tag):
                s = node.name + '，' + xldictstr(node.attrs, item_delimit='，')
            # elif isinstance(node, (bs4.NavigableString, code4101py.stdlib.bs4.NavigableString)):
            elif isinstance(node, bs4.NavigableString):
                # s = 'NavigableString'
                s = shorten(str(node), 200)
                if not s.strip():
                    s = '<??>'
            else:
                s = '遇到特殊类型，' + str(node)
            return s

        s = dfs_base(self.node(), mystr=mystr, prefix=prefix, linenum=linenum, **kwargs)
        return s

    def treestruct_stat(self):
        """生成一个两个二维表的统计数据
            ls1, ls2 = treestruct_stat()
                ls1： 结点规律表
                ls2： 属性规律表
        count_tagname、check_tag的功能基本都可以被这个函数代替
        """

        def text(t):
            """ 考虑到结果一般都是存储到excel，所以会把无法存成gbk的字符串删掉
            另外控制了每个元素的长度上限
            """
            s = ensure_gbk(t)
            s = s[:100]
            return s

        def depth(t):
            """结点t的深度"""
            return len(tuple(t.parents))

        t = self.contents[0]
        # ls1 = [['element序号', '层级', '结构', '父结点', '当前结点', '属性值/字符串值', '直接子结点结构']]
        # ls2 = [['序号', 'element序号', '当前结点', '属性名', '属性值']]  #
        ls1 = []  # 这个重点是分析结点规律
        ls2 = []  # 这个重点是分析属性规律
        i = 1
        while t:
            # 1 结点规律表
            d = depth(t)
            line = [i, d, '_' * d + str(d), tag_name(t.parent), tag_name(t),
                    text(xldictstr(t.attrs) if t.name else t),  # 结点存属性，字符串存值
                    subtag_names(t)]
            ls1.append(line)
            # 2 属性规律表
            if t.name:
                k = len(ls2)
                for attr, value in t.attrs.items():
                    ls2.append([k, i, tag_name(t), attr, value])
                    k += 1
            # 下个结点
            t = t.next_element
            i += 1
        df1 = pd.DataFrame.from_records(ls1, columns=['element序号', '层级', '结构', '父结点', '当前结点', '属性值/字符串值', '直接子结点结构'])
        df2 = pd.DataFrame.from_records(ls2, columns=['序号', 'element序号', '当前结点', '属性名', '属性值'])
        return df1, df2

    def count_tagname(self):
        """统计每个标签出现的次数：
             1                    w:rpr  650
             2                 w:rfonts  650
             3                   w:szcs  618
             4                      w:r  565
             5                     None  532
             6                      w:t  531
        """
        ct = collections.Counter()

        def inner(node):
            try:
                ct[node.name] += 1
                for t in node.children:
                    inner(t)
            except AttributeError:
                pass

        inner(self.node())
        return ct.most_common()

    def check_tag(self, tagname=None):
        """统计每个标签在不同层级出现的次数：

        :param tagname:
            None：统计全文出现的各种标签在不同层级出现次数
            't'等值： tagname参数允许只检查特殊标签情况，此时会将所有tagname设为第0级

        TODO 检查一个标签内部是否有同名标签？
        """
        d = defaultdict()

        def add(name, depth):
            if name not in d:
                d[name] = defaultdict(int)
            d[name][depth] += 1

        def inner(node, depth):
            if isinstance(node, bs4.ProcessingInstruction):
                add('ProcessingInstruction', depth)
            elif isinstance(node, bs4.Tag):
                if node.name == tagname and depth:
                    dprint(node, depth)  # tagname里有同名子标签
                add(node.name, depth)
                for t in node.children:
                    inner(t, depth + 1)
            elif isinstance(node, bs4.NavigableString):
                add('NavigableString', depth)
            else:
                add('其他特殊结点', depth)

        # 1 统计结点在每一层出现的次数
        if tagname:
            for t in self.node().find_all(tagname):
                inner(t, 0)
        else:
            inner(self.node(), 0)

        # 2 总出现次数和？

        return d

    def check_namespace(self):
        """检查名称空间问题，会同时检查标签名和属性名：
            1  cNvPr  pic:cNvPr(579)，wps:cNvPr(52)，wpg:cNvPr(15)
            2   spPr                   pic:spPr(579)，wps:spPr(52)
        """
        # 1 获得所有名称
        #    因为是采用node的原始xml文本，所以能保证会取得带有名称空间的文本内容
        ct0 = Counter(re.findall(r'<([a-zA-Z:]+)', str(self.node())))
        ct = defaultdict(str)
        s = set()
        for key, value in ct0.items():
            k = re.sub(r'.*:', '', key)
            if k in ct:
                s.add(k)
                ct[k] += f'，{key}({value})'
            else:
                ct[k] = f'{key}({value})'

        # 2 对有重复和无重复的元素划分存储
        ls1 = []  # 有重复的存储到ls1
        ls2 = []  # 没有重复的正常结果存储到ls2，可以不显示
        for k, v in ct.items():
            if k in s:
                ls1.append([k, v])
            else:
                ls2.append([k, v])

        # 3 显示有重复的情况
        # browser(ls1, filename='检查名称空间问题')
        return ls1


class MyBs4(BeautifulSoup, XmlParser):
    """xml、html 等数据通用处理算法，常用功能有：

    show_brief：显示xml结构
    count_tagname： 统计各个结点名称出现次数
    """

    def __init__(self, markup="", features='lxml', *args, **kwargs):
        # markup = Path(markup).read()
        # TODO: **kwargs我不知道怎么传进来啊，不过感觉也不删大雅没什么鸟用吧~~
        super().__init__(markup, features, *args, **kwargs)

    def insert_after(self, successor):
        pass

    def insert_before(self, successor):
        pass


____section_temp = """
"""


def mathjax_html_head(s):
    """增加mathjax解析脚本"""
    head = r"""<!DOCTYPE html>
<html>
<head>
<head><meta http-equiv=Content-Type content="text/html;charset=utf-8"></head>
<script src="https://a.cdn.histudy.com/lib/config/mathjax_config-klxx.js?v=1.1"></script>
<script type="text/javascript" async src="https://a.cdn.histudy.com/lib/mathjax/2.7.1/MathJax/MathJax.js?config=TeX-AMS-MML_SVG">
MathJax.Hub.Config(MATHJAX_KLXX_CONFIG);
</script>
</head>
<body>"""
    tail = '</body></html>'
    return head + s + tail


def 自动制作网页标题的导航栏(html_content, title='temphtml'):
    """
    :param html_content: 原始网页的完整内容
    :param title: 页面标题，默认会先找head/title，如果没有，则取一个随机名称（TODO 未实装，目前固定名称'test'）

    算法基本原理：读取原网页，找出所有h标签，并增设a锚点
        另外生成一个导航html文件
        然后再生成一个主文件，让用户通过主文件来浏览页面

    # 读取csdn博客并展示目录 （不过因为这个存在跳级，效果不是那么好）
    >> file = 自动制作网页标题的导航栏(requests.get(r'https://blog.csdn.net/code4101/article/details/83009000').content.decode('utf8'))
    >> browser(str(file))
    http://i2.tiimg.com/582188/64f40d235705de69.png
    """
    # 1 对原html，设置锚点，生成一个新的文件f2；生成导航目录文件f1。
    cnt = 0

    # TODO 目前不支持跳级的情况
    # 这个refs是可以用py算法生成的，目前是存储在github上引用
    refs = ['<html><head>',
            '<link rel=Stylesheet type="text/css" media=all href="https://code4101.github.io/css/navigation0.css">',
            '</head><body>']

    f2 = File(title + '_内容', Dir.TEMP, suffix='.html')

    def func(m):
        nonlocal cnt
        cnt += 1
        name, content = m.group('name'), m.group('inner')
        content = BeautifulSoup(content, 'lxml').get_text()
        refs.append(f'<a href="{f2}#生成导航栏浏览网页{cnt}" target="showframe"><{name}>{content}</{name}></a>')
        return f'<a name="生成导航栏浏览网页{cnt}"/>' + m.group()

    html_content = re.sub(r'<(?P<name>h\d+)(?:>|\s.*?>)(?P<body>\s*(?P<inner>.*?)\s*)</\1>',
                          func, html_content, flags=re.DOTALL)

    refs.append('</body>\n</html>')

    f1 = File(title + '_导航', Dir.TEMP, suffix='.html').write('\n'.join(refs), if_exists='replace')
    f2 = f2.write(html_content, if_exists='replace')

    # 2 生成首页 f0
    main_content = f"""<html>
<frameset cols="20%,80%">
	<frame src="{f1}">
	<frame src="{f2}" name="showframe">
</frameset></html>"""

    f0 = File(title, Dir.TEMP, suffix='.html').write(main_content, if_exists='replace')
    return f0



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# @Author : 陈坤泽
# @Email  : 877362867@qq.com
# @Date   : 2020/06/02 16:06

import concurrent.futures
import math
import os
import pprint
import re
import tempfile
import pathlib

import cv2
import numpy as np

import subprocess

try:
    import fitz
except ModuleNotFoundError:
    subprocess.run(['pip3', 'install', 'PyMuPdf'])
    import fitz

from pyxllib.algo.pupil import get_number_width
from pyxllib.file.specialist import File, Dir, writefile, filescopy, filesdel
from pyxllib.debug.pupil import dprint
from pyxllib.debug.specialist import browser
from pyxllib.cv.expert import imwrite, cv2pil
from pyxllib.cv.imfile import zoomsvg


class FitzPdf:
    def __init__(self, file):
        self.src_file = File(file)
        self.doc = fitz.open(str(file))

    def to_images(self, dst_dir=None, file_fmt='{filestem}_{number}.png', num_width=None, *,
                  scale=1, start=1, fmt_onepage=False):
        """ 将pdf转为若干页图片

        :param dst_dir: 目标目录
            默认情况下，只有一页pdf则存储到对应的pdf目录，多页则存储到同名子目录下
            如果不想这样被智能控制，只要指定明确的dst即可
        :param file_fmt: 后缀格式，包括修改导出的图片类型，注意要用 {} 占位符表示页码编号
        :param num_width: 生成的每一页文件编号，使用的数字前导0域宽
            默认根据pdf总页数来设置对应所用域宽
            0表示不设域宽
        :param scale: 对每页图片进行缩放
        :param start: 起始页码
        :param fmt_onepage: 当pdf就只有一页的时候，是否还对导出的图片编号
            默认只有一页的时候，进行优化，不增设后缀格式

        注：如果要导出单张图，可以用 FitzPdfPage.get_cv_image
        """
        # 1 基本参数计算
        srcfile, doc = self.src_file, self.doc
        filestem, n_page = srcfile.stem, doc.pageCount

        # 自动推导目标目录
        if dst_dir is None:
            dst_dir = Dir(srcfile.stem, srcfile.parent) if n_page > 1 else Dir(srcfile.parent)
        Dir(dst_dir).ensure_dir()

        # 域宽
        num_width = num_width or get_number_width(n_page)  # 根据总页数计算需要的对齐域宽

        # 2 导出图片
        if fmt_onepage or n_page != 1:  # 多页的处理规则
            for i in range(n_page):
                im = self.get_page(i).get_cv_image(scale)
                number = ('{:0' + str(num_width) + 'd}').format(i + start)  # 前面的括号不要删，这样才是完整的一个字符串来使用format
                imwrite(im, File(file_fmt.format(filestem=filestem, number=number), dst_dir))
        else:
            im = self.get_page(0).get_cv_image(scale)
            imwrite(im, File(srcfile.stem + os.path.splitext(file_fmt)[1], dst_dir))

    def get_page(self, number):
        return FitzPdfPage(self.doc.loadPage(number))


class FitzPdfPage:
    def __init__(self, page):
        self.page = page

    def get_svg_image(self, scale=1):
        # svg 是一段表述性文本
        txt = self.page.getSVGimage()
        if scale != 1:
            txt = zoomsvg(txt, scale)
        return txt

    def _get_png_data(self, scale=1):
        # TODO 增加透明通道？
        if scale != 1:
            pix = self.page.getPixmap(fitz.Matrix(scale, scale))  # 长宽放大到scale倍
        else:
            pix = self.page.getPixmap()
        return pix.getPNGData()

    def get_cv_image(self, scale=1):
        arr = np.fromstring(self._get_png_data(scale), dtype=np.uint8)
        return cv2.imdecode(arr, flags=1)

    def get_pil_image(self, scale=1):
        # TODO 可以优化，直接从内存数据转pil，不用这样先转cv再转pil
        return cv2pil(self.get_cv_image(scale))

    def write_image(self, outfile, *, scale=1, if_exists=None):
        """ 转成为文件 """
        f = File(outfile)
        suffix = f.suffix.lower()

        if suffix == '.svg':
            content = self.get_svg_image()
            f.write(content, if_exists=if_exists)
        else:
            im = self.get_cv_image(scale)
            imwrite(im, if_exists=if_exists)

    def get_text(self, fmt='text'):
        """
        :param fmt: 存储格式，可以获得整页的纯文本，也可以获得dict结构存储的内容
        """
        return self.page.getText(fmt)


class DemoFitz:
    """
    安装： pip install PyMuPdf
    使用： import fitz
    官方文档： https://pymupdf.readthedocs.io/en/latest/intro/
        demo： https://github.com/rk700/PyMuPDF/tree/master/demo
        examples： https://github.com/rk700/PyMuPDF/tree/master/examples
    """

    def __init__(self, file):
        self.doc = fitz.open(file)

    def message(self):
        """查看pdf文档一些基础信息"""
        dprint(fitz.version)  # fitz模块的版本
        dprint(self.doc.pageCount)  # pdf页数
        dprint(self.doc._getXrefLength())  # 文档的对象总数

    def getToC(self):
        """获得书签目录"""
        toc = self.doc.getToC()
        browser(toc)

    def setToC(self):
        """设置书签目录
        可以调层级、改名称、修改指向页码
        """
        toc = self.doc.getToC()
        toc[1][1] = '改标题名称'
        self.doc.setToC(toc)
        file = File('a.pdf', Dir.TEMP).to_str()
        self.doc.save(file, garbage=4)
        browser(file)

    def setToC2(self):
        """修改人教版教材的标签名"""
        toc = self.doc.getToC()
        newtoc = []
        for i in range(len(toc)):
            name = toc[i][1]
            if '.' in name: continue
            # m = re.search(r'\d+', name)
            # if m: name = name.replace(m.group(), digits2chinese(int(m.group())))
            m = re.search(r'([一二三四五六]年级).*?([上下])', name)
            if i < len(toc) - 1:
                pages = toc[i + 1][2] - toc[i][2] + 1
            else:
                pages = self.doc.pageCount - toc[i][2] + 1
            toc[i][1] = m.group(1) + m.group(2) + '，' + str(pages)
            newtoc.append(toc[i])
        self.doc.setToC(newtoc)
        file = writefile(b'', 'a.pdf', if_exists='replace')
        self.doc.save(file, garbage=4)

    def rearrange_pages(self):
        """重新布局页面"""
        self.doc.select([0, 0, 1])  # 第1页展示两次后，再跟第2页
        file = writefile(b'', 'a.pdf', root=Dir.TEMP, if_exists='replace')
        self.doc.save(file, garbage=4)  # 注意要设置garbage，否则文档并没有实际删除内容压缩文件大小
        browser(file)

    def page2png(self, page=0):
        """ 查看单页渲染图片 """
        page = self.doc.loadPage(page)  # 索引第i页，下标规律同py，支持-1索引最后页
        # dprint(page.bound())  # 页面边界，x,y轴同图像处理中的常识定义，返回Rect(x0, y0, x1, y1)

        pix = page.getPixmap(fitz.Matrix(2, 2))  # 获得页面的RGBA图像，Pixmap类型；还可以用page.getSVGimage()获得矢量图
        # pix.writePNG('page-0.png')  # 将Pixmal
        pngdata = pix.getPNGData()  # 获png文件的bytes字节码
        # print(len(pngdata))
        # browser(pngdata, 'a.png')  # 用我的工具函数打开图片

        return pngdata

    def pagetext(self):
        """单页上的文本"""
        page = self.doc[0]

        # 获得页面上的所有文本，还支持参数： html，dict，xml，xhtml，json
        text = page.getText('text')
        dprint(text)

        # 获得页面上的所有文本（返回字典对象）
        textdict = page.getText('dict')
        textdict['blocks'] = textdict['blocks'][:-1]
        browser(pprint.pformat(textdict))

    def text(self):
        """获得整份pdf的所有文本"""
        return '\n'.join([page.getText('text') for page in self.doc])

    def xrefstr(self):
        """查看pdf文档的所有对象"""
        xrefstr = []
        n = self.doc._getXrefLength()
        for i in range(1, n):  # 注意下标实际要从1卡开始
            # 可以边遍历边删除，不影响下标位置，因为其本质只是去除关联引用而已
            xrefstr.append(self.doc._getXrefString(i))
        browser('\n'.join(xrefstr))

    def page_add_ele(self):
        """往页面添加元素
        添加元素前后xrefstr的区别： https://paste.ubuntu.com/p/Dxhnzp4XJ2/
        """
        self.doc.select([0])
        page = self.doc.loadPage(0)
        # page.insertText(fitz.Point(100, 200), 'test\ntest')
        file = File('a.pdf', Dir.TEMP).to_str()
        dprint(file)
        self.doc.save(file, garbage=4)
        browser(file)


__backup = """
下面这些代码先留着不要删，以后做旧代码兼容分析，升级的时候，需要参考

这里有些功能是编校还是不断使用的，但接口等需要重新优化
"""

# def pdf2svg_oldversion(pdffile, target=None, *, trim=False):
#     """新版的，pymupdf生成的svg无法配合inkscape进行trim，所以这个旧版暂时还是要保留
#
#     :param pdffile: 一份pdf文件
#     :param target: 目标目录
#         None：
#             如果只有一页，转换到对应目录下同名svg文件
#             如果有多页，转换到同名目录下多份svg文件
#     :param trim:
#         True: 去除边缘空白
#     :return:
#
#     需要第三方工具：pdf2svg（用于文件格式转换），inkscape（用于svg编辑优化）
#         注意pdf2svg的参数不支持中文名，因为这个功能限制，搞得我这个函数实现好麻烦！
#         还要在临时文件夹建立文件，因为重名文件+多线程问题，还曾引发一个bug搞了一下午。
#         （这些软件都以绿色版形式整理在win3/imgtools里）
#
#     注意！！！ 这个版本的代码先不要删！先不要删！先不要删！包括pdf2svg.exe那个蠢货软件也先别删！
#         后续研究inkscape这个蠢货的-D参数在处理pymupdf生成的svg为什么没用的时候可以进行对比
#     """
#     import fitz
#     pages = fitz.open(pdffile).pageCount
#
#     basename = tempfile.mktemp()
#     f1 = basename + '.pdf'
#     filescopy(pdffile, f1)  # 复制到临时文件，防止中文名pdf2svg工具处理不了
#
#     if pages == 1:
#         if target is None: target = pdffile[:-3] + 'svg'
#         f2 = basename + '.svg'
#         # print(['pdf2svg.exe', f1, f2])
#         subprocess.run(['pdf2svg.exe', f1, f2])
#
#         if trim: subprocess.run(['inkscape.exe', '-f', f2, '-D', '-l', f2])
#         filescopy(f2, target)
#     else:
#         if target is None: target = pdffile[:-4] + '_svg\\'
#         executor = concurrent.futures.ThreadPoolExecutor()
#         File(basename + '/').ensure_parent()
#
#         def func(f1, f2, i):
#             subprocess.run(['pdf2svg.exe', f1, f2, str(i)])
#             if trim: subprocess.run(['inkscape.exe', '-f', f2, '-D', '-l', f2])
#             filescopy(f2, target + f'{i}.svg')
#
#         for i in range(1, pages + 1):
#             f2 = basename + f'\\{i}.svg'
#             executor.submit(func, f1, f2, i)
#         executor.shutdown()
#         filescopy(basename, target[:-1])
#         filesdel(basename + '/')
#
#     filesdel(f1)


# def pdf2imagebase(pdffile, target=None, *, ext, scale=None, if_exists=None):
#     """
#     使用python的PyMuPdf模块，不需要额外插件
#     导出的图片从1开始编号
#     TODO 要加多线程？效率影响大吗？
#
#     :param pdffile: pdf原文件
#     :type target: 相对于原文件所在目录的目标目录名，也可以写文件名，表示重命名
#         None：
#             当该pdf只有1页时，才默认把图片转换到当前目录。
#             否则默认新建一个文件夹来存储图片。（目录名默认为文件名）
#     :param scale: 缩放尺寸
#         1：原尺寸
#         1.5：放大为原来的1.5倍
#     :param ext: 导出的图片格式
#     :return: 返回生成的图片列表
#     """
#
#     import fitz
#     # 1 基本参数计算
#     pdffile = File(pdffile)
#     pdf = fitz.open(pdffile)
#     num_pages = pdf.pageCount
#
#     # 大于1页的时候，默认新建一个文件夹来存储图片
#     if target is None:
#         if num_pages > 1:
#             target = File(pdffile).stem + '/'
#         else:
#             target = File(pdffile).dirname + '/'
#
#     newfile = str(pathlib.Path(pdffile.parent / target))
#     if newfile.endswith('.pdf'):
#         newfile = os.path.splitext(newfile)[0] + ext
#     Dir(os.path.dirname(newfile)).ensure_parent()
#
#     # 2 分析导出的图片文件名
#     files = []
#     if num_pages == 1:
#         FitzPdfPage(pdf.loadPage(0)).write_image(newfile, if_exists=if_exists)
#         files.append(newfile)
#     else:  # 有多页
#         number_width = get_number_width(num_pages)  # 根据总页数计算需要的对齐域宽
#         stem, ext = os.path.splitext(newfile)
#         for i in range(num_pages):
#             name = ('-{:0' + str(number_width) + 'd}').format(i + 1)  # 前面的括号不要删，这样才是完整的一个字符串来使用format
#             file = stem + name + ext
#             files.append(file)
#             FitzPdfPage(pdf.loadPage(i)).write_image(file, if_exists=if_exists)
#     return files


# def pdf2png(pdffile, target=None, scale=None):
#     """
#     :param pdffile: pdf路径
#     :param target: 目标位置
#     :param scale: 缩放比例
#     :return: list，生成的png图片清单
#
#     # 可以不写target，默认处理：如果单张png则在同目录，多张则会建个同名目录存储
#     >> pdf2png(r'D:\slns+\immovables\immovables_data\test\X\A0001.pdf')
#
#     # 指定存放位置：
#     >> pdf2png(r'D:\slns+\immovables\immovables_data\test\X\A0001.pdf', r'D:\slns+\immovables\immovables_data\test\X')
#
#     """
#     return pdf2imagebase(pdffile, target=target, scale=scale, ext='.png')


# def pdf2jpg(pdffile, target=None, scale=None):
#     return pdf2imagebase(pdffile, target=target, scale=scale, ext='.jpg')
#


# def pdf2svg(pdffile, target=None, scale=None, trim=False):
#     """
#     :param pdffile: 见pdf2imagebase
#     :param target: 见pdf2imagebase
#     :param scale: 见pdf2imagebase
#     :param trim: 如果使用裁剪功能，会调用pdf-crop-margins第三方工具
#         https://pypi.org/project/pdfCropMargins/
#     :return:
#     """
#     if trim:  # 先对pdf文件进行裁剪再转换
#         pdf = File(pdffile)
#         newfile = File('origin.pdf', pdf.parent).to_str()
#         pdf.copy(newfile)
#         # subprocess.run(['pdf-crop-margins.exe', '-p', '0', newfile, '-o', pdffile], stderr=subprocess.PIPE) # 本少： 会裁过头！
#         # 本少： 对于上下边处的 [] 分数等，会裁过头，先按百分比 -p 0 不留边，再按绝对点数收缩/扩张 -a -1  负数为扩张，单位为bp
#         # 本少被自己坑了，RamDisk 与 pdf-crop-margins.exe 配合，只能取 SCSI 硬盘，如果 Direct-IO 就不行，还不报错，还以为是泽少写的代码连报错都不会
#         subprocess.run(['pdf-crop-margins.exe', '-p', '0', '-a', '-1', newfile, '-o', pdffile],
#                        stderr=subprocess.PIPE)
#     # TODO 有时丢图
#     pdf2imagebase(pdffile, target=target, scale=scale, ext='.svg')


# def pdfs2pngs(src, scale=None, pinterval=None):
#     """ 将目录下所有pdf转png
#
#     :param src: 原pdf数据路径
#     :param scale: 转图片时缩放比例，例如2表示长宽放大至2被
#     :param pinterval: 每隔多少个pdf输出处理进度
#         默认None，不输出
#     """
#     from functools import partial
#     func = partial(pdf2png, scale=scale)
#     Dir(src).select('**/*.pdf').procpaths(func, pinterval=pinterval)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# @Author : 陈坤泽
# @Email  : 877362867@qq.com
# @Date   : 2021/06/04 09:14


from pyxllib.xlai import *


def test_labeldata():
    os.chdir(r'D:\home\datasets\RealEstate2020\temp')

    # 1 初始化
    ld = LabelmeDataset(r'D:\home\datasets\RealEstate2020\temp')
    # 升级字典
    # ld.to_labelattrs()

    # 2 转coco
    gt_dict = ld.to_coco_gt_dict()
    # pprint.pprint(gt_dict)

    # 3 转表格
    # ld.to_excel('test.xlsx')

    # 4 coco对象
    cgd = CocoGtData(gt_dict)

    # 5 coco对象再转回labelme，礼尚往来
    ld2 = cgd.to_labelme(r'D:\home\datasets\RealEstate2020\temp2')
    ld2.writes()

    # 6 ld2再转回coco，梅开二度
    gt_dict = ld2.to_coco_gt_dict()
    pprint.pprint(gt_dict)


def test_tqdm():
    """ 下载情况下的进度显示怎么做 """
    with tqdm(unit="B", unit_scale=True, miniters=1, desc='a.test', leave=True) as t:
        t.total = 10000000
        for i in range(10):
            t.update(1000000)  # type: ignore
            time.sleep(1)


class ArithmeticPerf(PerfTest):
    """ 可能数值比较小，这个实验不太明显吧

    >> ArithmeticPerf().perf(number=10000, repeat=100)
    add 用时(秒) 总和: 1.266	均值标准差: 0.127±0.019	总数: 10	最小值: 0.094	最大值: 0.167 运行结果：121
    div 用时(秒) 总和: 1.281	均值标准差: 0.128±0.017	总数: 10	最小值: 0.097	最大值: 0.150 运行结果：1.75
    mul 用时(秒) 总和: 1.121	均值标准差: 0.112±0.012	总数: 10	最小值: 0.095	最大值: 0.131 运行结果：3388
    sub 用时(秒) 总和: 1.214	均值标准差: 0.121±0.015	总数: 10	最小值: 0.102	最大值: 0.150 运行结果：33
    """

    def perf_add(self):
        return 77 + 44

    def perf_sub(self):
        return 77 - 44

    def perf_mul(self):
        return 77 * 44

    def perf_div(self):
        return 77 / 44


class CallPerf(PerfTest):
    """ 奕本纠结嵌套太多产生的效率问题

    >> CallPerf().perf(number=100000, repeat=100)
    func1 用时(秒) 总和: 7.947	均值标准差: 0.079±0.019	总数: 100	最小值: 0.063	最大值: 0.172 运行结果：bc
    func2 用时(秒) 总和: 12.581	均值标准差: 0.126±0.038	总数: 100	最小值: 0.093	最大值: 0.268 运行结果：bc
    """

    def perf_func1(self):
        def a():
            def b():
                def c():
                    return 'abc'.replace('a', '')

                return c()

            return b()

        return a()

    def perf_func2(self):
        return re.sub(r'a', '', 'abc')


def 文件结构(dirpath, depth=0):
    from humanfriendly import format_size

    d = Dir(dirpath)
    print('\t' * depth + f'☐{d.name}, {format_size(d.size, binary=True)}')
    for dd in d.select_dirs('*'):
        # 不处理的一些目录
        if dd.name in ('__pycache__',):
            continue
        文件结构(dd, depth + 1)
    for f in d.select_files('*'):
        print('\t' * (depth + 1) + f'{f.name}, {format_size(f.size, binary=True)}')


if __name__ == '__main__':
    with TicToc(__name__):
        # test_labeldata()

        # 文件结构(r'C:\ProgramData\Anaconda3\Lib\site-packages\iopath')
        # 文件结构(r'D:\slns\detectron2\detectron2')

        pass



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# @Author : 陈坤泽
# @Email  : 877362867@qq.com
# @Date   : 2021/06/06 10:51


def vector_compare(x, y):
    """
    :param x: 一个数值向量
    :param y: 一个数值向量，理论上长度要跟x相同
    :return: 返回<、=、>、?
        =，各个位置的数值都相同
        <，x各个位置的值都≤y，且至少有一处值<y
        >，x各个位置的值都≥，且至少有一处值>y
        ?，其他情况，x、y各有优劣

    >>> vector_compare([1, 2, 3], [1, 3, 3])
    '<'
    >>> vector_compare([1, 2, 3], [1, 1, 4])
    '?'
    >>> vector_compare({'a': 1, 'b': 2}, {'a': 1, 'b': 2, 'c': 0})
    '='
    >>> vector_compare({'b': 3, 'a': 1}, {'a': 1, 'b': 2, 'c': 0})
    '>'
    >>> vector_compare({'b': 3, 'a': 1}, {'a': 1, 'b': 2, 'c': 1})
    '?'
    """
    # 1 不同数据类型处理
    if isinstance(x, dict) and isinstance(y, dict):
        # 字典比较比较特别，要先做个映射
        keys = set(x.keys()) | set(y.keys())
        a = [x.get(k, 0) - y.get(k, 0) for k in keys]
    else:
        a = [i - j for i, j in zip(x, y)]

    # 2 算法
    n, m = max(a), min(a)
    if n == m == 0:
        return '='
    elif n <= 0:
        return '<'
    elif m >= 0:
        return '>'
    else:
        return '?'



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# @Author : 陈坤泽
# @Email  : 877362867@qq.com
# @Date   : 2020/09/30 10:45

"""
这个文件默认不加载到cv里，需要使用的时候再导入

from pyxlib.cv.debugtools import *
"""

import math

import cv2
import numpy as np

from pyxllib.time.specialist import TicToc
from pyxllib.cv.expert import CvPlot, imread


class TrackbarTool:
    """ 滑动条控件组
    """

    def __init__(self, winname, img, flags=0):
        img = imread(img)
        cv2.namedWindow(winname, flags)
        cv2.imshow(winname, img)
        self.winname = winname
        self.img = img
        self.trackbar_names = {}

    def imshow(self, img=None):
        """ 刷新显示的图片 """
        if img is None:
            img = self.img
        cv2.imshow(self.winname, img)

    def default_run(self, x):
        """ 默认执行器，这个在类继承后，基本都是要自定义成自己的功能的

        TODO 从1滑到20，会运行20次，可以研究一个机制，来只运行一次
        """
        kwargs = {}
        for k in self.trackbar_names.keys():
            kwargs[k] = self[k]
        print(kwargs)

    def create_trackbar(self, trackbar_name, count, value=0, on_change=None):
        """ 创建一个滑动条

        :param trackbar_name: 滑动条名称
        :param count: 上限值
        :param on_change: 回调函数
        :param value: 初始值
        :return:
        """
        if on_change is None:
            on_change = self.default_run
        cv2.createTrackbar(trackbar_name, self.winname, value, count, on_change)

    def __getitem__(self, item):
        """ 可以通过 Trackbars 来获取滑动条当前值

        :param item: 滑动条名称
        :return: 当前取值
        """
        return cv2.getTrackbarPos(item, self.winname)


class ErodeTool(TrackbarTool):
    r""" 腐蚀 erode

    HoughLinesPTool('handwriting/C100001448573-002.png')
    """

    def __init__(self, img, winname='ErodeTool', flags=1):
        super().__init__(winname, img, flags)

        # 1 增加控件
        self.create_trackbar('size_x', 100, 15)
        self.create_trackbar('size_y', 100, 15)

        # 2 初始化执行一遍
        self.default_run(0)

    def default_run(self, x):
        """ 默认执行器
        """
        tt = TicToc()
        size = self['size_x'], self['size_y']
        element = cv2.getStructuringElement(cv2.MORPH_RECT, size)
        dst = cv2.erode(self.img, element)
        self.imshow(dst)
        tt.toc(f'element_size={size}')


class HoughLinesPTool(TrackbarTool):
    r""" 霍夫线段检测

    HoughLinesPTool('handwriting/C100001448573-002.png')
    """

    def __init__(self, img, winname='HoughLinesPTool', flags=1):
        super().__init__(winname, img, flags)

        # 1 增加控件
        # 以像素为单位的距离精度
        self.create_trackbar('rho*10', 100, 10)
        # 以弧度为单位的角度精度
        self.create_trackbar('theta*pi/360', 10, 2)
        # 累加平面的阈值参数
        self.create_trackbar('threshold', 200, 80)
        # 最低线段长度
        self.create_trackbar('minLineLength', 200, 50)  # 如果要显示全，这里名称最好只有10个字符，也可能是跟图片尺寸有关，有待进一步观察
        # 允许将同一行点与点之间连接起来的最大的距离
        self.create_trackbar('maxLineGap', 100, 10)

        # 2 图片预处理，这里暂时用自适应二值化，后面可以把该部分也变成可调试项
        if self.img.ndim == 3:
            gray_img = cv2.cvtColor(self.img, cv2.COLOR_BGR2GRAY)
        else:
            gray_img = self.img
        self.binary_img = cv2.adaptiveThreshold(gray_img, 255, 0, 1, 11, 3)

        # 3 初始化执行一遍
        self.default_run(0)

    def default_run(self, x):
        """ 默认执行器
        """
        tt = TicToc()
        lines = cv2.HoughLinesP(self.binary_img,
                                self['rho*10'] / 10, self['theta*pi/360'] * math.pi / 360,
                                self['threshold'], self['minLineLength'], self['maxLineGap'])
        if lines is None: lines = np.array([])
        # 处理用二值化，显示用原图
        self.imshow(CvPlot.lines(self.img, lines))
        tt.toc(f'x={x} lines.shape={lines.squeeze().shape}')
